from typing import Any, Dict, Iterable
from collections.abc import MutableMapping

from ._value import AbstractBaseMutableTag, AbstractBaseTag
from ._dtype import AnyNBT
from ._int import ByteTag, ShortTag, IntTag, LongTag
from ._float import FloatTag, DoubleTag
from ._string import StringTag
from ._list import ListTag
from ._array import ByteArrayTag, IntArrayTag, LongArrayTag

class CompoundTag(AbstractBaseMutableTag, MutableMapping[str, AnyNBT]):
    tag_id: int
    def __init__(self, value: Any = (), **kwvals: AnyNBT): ...
    @staticmethod
    def fromkeys(keys: Iterable[str], value: AbstractBaseTag = None): ...
    @property
    def py_dict(self) -> Dict[str, AnyNBT]: ...
    def get_byte(self, key: str, default: ByteTag = None) -> ByteTag: ...
    def get_short(self, key: str, default: ShortTag = None) -> ShortTag: ...
    def get_int(self, key: str, default: IntTag = None) -> IntTag: ...
    def get_long(self, key: str, default: LongTag = None) -> LongTag: ...
    def get_float(self, key: str, default: FloatTag = None) -> FloatTag: ...
    def get_double(self, key: str, default: DoubleTag = None) -> DoubleTag: ...
    def get_string(self, key: str, default: StringTag = None) -> StringTag: ...
    def get_list(self, key: str, default: ListTag = None) -> ListTag: ...
    def get_compound(self, key: str, default: CompoundTag = None) -> CompoundTag: ...
    def get_byte_array(
        self, key: str, default: ByteArrayTag = None
    ) -> ByteArrayTag: ...
    def get_int_array(self, key: str, default: IntArrayTag = None) -> IntArrayTag: ...
    def get_long_array(
        self, key: str, default: LongArrayTag = None
    ) -> LongArrayTag: ...
    def setdefault_byte(self, key: str, default: ByteTag = None) -> ByteTag: ...
    def setdefault_short(self, key: str, default: ShortTag = None) -> ShortTag: ...
    def setdefault_int(self, key: str, default: IntTag = None) -> IntTag: ...
    def setdefault_long(self, key: str, default: LongTag = None) -> LongTag: ...
    def setdefault_float(self, key: str, default: FloatTag = None) -> FloatTag: ...
    def setdefault_double(self, key: str, default: DoubleTag = None) -> DoubleTag: ...
    def setdefault_string(self, key: str, default: StringTag = None) -> StringTag: ...
    def setdefault_list(self, key: str, default: ListTag = None) -> ListTag: ...
    def setdefault_compound(
        self, key: str, default: CompoundTag = None
    ) -> CompoundTag: ...
    def setdefault_byte_array(
        self, key: str, default: ByteArrayTag = None
    ) -> ByteArrayTag: ...
    def setdefault_int_array(
        self, key: str, default: IntArrayTag = None
    ) -> IntArrayTag: ...
    def setdefault_long_array(
        self, key: str, default: LongArrayTag = None
    ) -> LongArrayTag: ...



from typing import Iterable, overload, List, TypeVar
from collections.abc import MutableSequence

from ._value import AbstractBaseMutableTag
from ._dtype import AnyNBT
from ._int import ByteTag, ShortTag, IntTag, LongTag
from ._float import FloatTag, DoubleTag
from ._string import StringTag
from ._compound import CompoundTag
from ._array import ByteArrayTag, IntArrayTag, LongArrayTag

AnyNBTT = TypeVar("AnyNBTT", bound=AnyNBT)

class ListTag(AbstractBaseMutableTag, MutableSequence[AnyNBTT]):
    tag_id: int
    list_data_type: int
    @overload
    def __init__(self, value: Iterable[AnyNBTT] = ()) -> None: ...
    @overload
    def __init__(self, value: Iterable[AnyNBTT] = (), list_data_type=1) -> None: ...
    @property
    def py_list(self) -> List[AnyNBTT]: ...
    def copy(self) -> ListTag[AnyNBTT]: ...
    def get_byte(self, index: int) -> ByteTag: ...
    def get_short(self, index: int) -> ShortTag: ...
    def get_int(self, index: int) -> IntTag: ...
    def get_long(self, index: int) -> LongTag: ...
    def get_float(self, index: int) -> FloatTag: ...
    def get_double(self, index: int) -> DoubleTag: ...
    def get_string(self, index: int) -> StringTag: ...
    def get_list(self, index: int) -> ListTag: ...
    def get_compound(self, index: int) -> CompoundTag: ...
    def get_byte_array(self, index: int) -> ByteArrayTag: ...
    def get_int_array(self, index: int) -> IntArrayTag: ...
    def get_long_array(self, index: int) -> LongArrayTag: ...



from typing import Union, Iterator, overload, Any
from ._value import AbstractBaseImmutableTag

class StringTag(AbstractBaseImmutableTag):
    @overload
    def __init__(self, value: Any = ""): ...
    def __contains__(self, o: str) -> bool:
        raise NotImplementedError
    def __eq__(self, x: object) -> bool:
        raise NotImplementedError
    def __getitem__(self, i: Union[int, slice]) -> str:
        raise NotImplementedError
    def __iter__(self) -> Iterator[str]:
        raise NotImplementedError
    def __len__(self) -> int:
        raise NotImplementedError
    @property
    def py_str(self) -> str: ...



## This file is generated by tempita. Do not modify this file directly or your changes will get overwritten.
## To edit this file edit the template in template/src

import numpy
cimport numpy
numpy.import_array()
from io import BytesIO
from copy import deepcopy
import warnings

from . import __major__
from ._value cimport AbstractBaseMutableTag
from ._const cimport CommaSpace, ID_BYTE_ARRAY, ID_INT_ARRAY, ID_LONG_ARRAY
from ._util cimport write_array, BufferContext, read_int, read_data
if __major__ <= 2:
    from ._util import primitive_conversion
from ._dtype import EncoderType


cdef class AbstractBaseArrayTag(AbstractBaseMutableTag):
    def __getitem__(AbstractBaseArrayTag self, item):
        raise NotImplementedError

    def __setitem__(AbstractBaseArrayTag self, key, value):
        raise NotImplementedError

    def __array__(AbstractBaseArrayTag self, dtype=None):
        raise NotImplementedError

    def __len__(AbstractBaseArrayTag self):
        raise NotImplementedError

    @property
    def np_array(AbstractBaseArrayTag self):
        """
        A numpy array holding the same internal data.
        Changes to the array will also modify the internal state.
        """
        raise NotImplementedError

    @property
    def py_data(self):
        """
        A python representation of the class. Note that the return type is undefined and may change in the future.
        You would be better off using the py_{type} or np_array properties if you require a fixed type.
        This is here for convenience to get a python representation under the same property name.
        """
        return self.np_array


cdef inline ByteArrayTag read_byte_array_tag(BufferContext buffer, bint little_endian):
    cdef ByteArrayTag tag = ByteArrayTag.__new__(ByteArrayTag)
    cdef int length = read_int(buffer, little_endian)
    cdef int byte_length = length * 1
    cdef char*arr = read_data(buffer, byte_length)
    data_type = numpy.dtype("int8") if little_endian else numpy.dtype("int8")
    tag.value_ = numpy.array(numpy.frombuffer(arr[:byte_length], dtype=data_type, count=length), numpy.dtype("int8")).ravel()
    return tag


cdef class ByteArrayTag(AbstractBaseArrayTag):
    """This class behaves like an 1D Numpy signed integer array with each value stored in a byte."""
    tag_id = ID_BYTE_ARRAY

    def __init__(ByteArrayTag self, object value = ()):
        self.value_ = numpy.asarray(value, numpy.dtype("int8")).ravel()

    cdef str _to_snbt(ByteArrayTag self):
        cdef long long elem
        cdef list tags = []
        for elem in self.value_:
            tags.append(f"{elem}B")
        return f"[B;{CommaSpace.join(tags)}]"

    cdef void write_payload(
        ByteArrayTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_array(
            numpy.asarray(self.value_, numpy.dtype("int8") if little_endian else numpy.dtype("int8")),
            buffer,
            1,
            little_endian
        )

    def __str__(ByteArrayTag self):
        return str(self.value_)

    def __reduce__(ByteArrayTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(ByteArrayTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(ByteArrayTag self):
        return self.__class__(self.value_)

    @property
    def np_array(ByteArrayTag self):
        """
        A numpy array holding the same internal data.
        Changes to the array will also modify the internal state.
        """
        return self.value_

    def __repr__(ByteArrayTag self):
        return f"{self.__class__.__name__}({list(self.value_)})"

    def __eq__(ByteArrayTag self, other):
        cdef ByteArrayTag other_
        if isinstance(other, ByteArrayTag):
            other_ = other
            return numpy.array_equal(self.value_, other_.value_)
        elif __major__ <= 2 and isinstance(other, AbstractBaseArrayTag):
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return numpy.array_equal(self.value_, primitive_conversion(other))
        return NotImplemented

    def __getitem__(ByteArrayTag self, item):
        return self.value_.__getitem__(item)

    def __setitem__(ByteArrayTag self, key, value):
        self.value_.__setitem__(key, value)

    def __array__(ByteArrayTag self, dtype=None):
        return numpy.asarray(self.value_, dtype=dtype)

    def __len__(ByteArrayTag self):
        return len(self.value_)

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) + primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) - primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) * primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __matmul__(self, other):
            warnings.warn(f"__matmul__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) @ primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) / primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) // primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) % primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo).astype(numpy.dtype("int8"))

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) << primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) >> primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) & primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) ^ primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) | primitive_conversion(other)).astype(numpy.dtype("int8"))

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) + primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) - primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) * primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) / primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) // primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) % primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo).astype(numpy.dtype("int8"))

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) << primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) >> primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) & primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) ^ primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) | primitive_conversion(self)).astype(numpy.dtype("int8"))

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (-self.value_).astype(numpy.dtype("int8"))

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (+self.value_).astype(numpy.dtype("int8"))

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on ByteArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_).astype(numpy.dtype("int8"))


cdef inline IntArrayTag read_int_array_tag(BufferContext buffer, bint little_endian):
    cdef IntArrayTag tag = IntArrayTag.__new__(IntArrayTag)
    cdef int length = read_int(buffer, little_endian)
    cdef int byte_length = length * 4
    cdef char*arr = read_data(buffer, byte_length)
    data_type = numpy.dtype("<i4") if little_endian else numpy.dtype(">i4")
    tag.value_ = numpy.array(numpy.frombuffer(arr[:byte_length], dtype=data_type, count=length), numpy.int32).ravel()
    return tag


cdef class IntArrayTag(AbstractBaseArrayTag):
    """This class behaves like an 1D Numpy signed integer array with each value stored in a int."""
    tag_id = ID_INT_ARRAY

    def __init__(IntArrayTag self, object value = ()):
        self.value_ = numpy.asarray(value, numpy.int32).ravel()

    cdef str _to_snbt(IntArrayTag self):
        cdef long long elem
        cdef list tags = []
        for elem in self.value_:
            tags.append(f"{elem}")
        return f"[I;{CommaSpace.join(tags)}]"

    cdef void write_payload(
        IntArrayTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_array(
            numpy.asarray(self.value_, numpy.dtype("<i4") if little_endian else numpy.dtype(">i4")),
            buffer,
            4,
            little_endian
        )

    def __str__(IntArrayTag self):
        return str(self.value_)

    def __reduce__(IntArrayTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(IntArrayTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(IntArrayTag self):
        return self.__class__(self.value_)

    @property
    def np_array(IntArrayTag self):
        """
        A numpy array holding the same internal data.
        Changes to the array will also modify the internal state.
        """
        return self.value_

    def __repr__(IntArrayTag self):
        return f"{self.__class__.__name__}({list(self.value_)})"

    def __eq__(IntArrayTag self, other):
        cdef IntArrayTag other_
        if isinstance(other, IntArrayTag):
            other_ = other
            return numpy.array_equal(self.value_, other_.value_)
        elif __major__ <= 2 and isinstance(other, AbstractBaseArrayTag):
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return numpy.array_equal(self.value_, primitive_conversion(other))
        return NotImplemented

    def __getitem__(IntArrayTag self, item):
        return self.value_.__getitem__(item)

    def __setitem__(IntArrayTag self, key, value):
        self.value_.__setitem__(key, value)

    def __array__(IntArrayTag self, dtype=None):
        return numpy.asarray(self.value_, dtype=dtype)

    def __len__(IntArrayTag self):
        return len(self.value_)

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) + primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) - primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) * primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __matmul__(self, other):
            warnings.warn(f"__matmul__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) @ primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) / primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) // primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) % primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo).astype(numpy.dtype(">i4"))

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) << primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) >> primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) & primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) ^ primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) | primitive_conversion(other)).astype(numpy.dtype(">i4"))

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) + primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) - primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) * primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) / primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) // primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) % primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo).astype(numpy.dtype(">i4"))

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) << primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) >> primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) & primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) ^ primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) | primitive_conversion(self)).astype(numpy.dtype(">i4"))

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (-self.value_).astype(numpy.dtype(">i4"))

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (+self.value_).astype(numpy.dtype(">i4"))

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on IntArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_).astype(numpy.dtype(">i4"))


cdef inline LongArrayTag read_long_array_tag(BufferContext buffer, bint little_endian):
    cdef LongArrayTag tag = LongArrayTag.__new__(LongArrayTag)
    cdef int length = read_int(buffer, little_endian)
    cdef int byte_length = length * 8
    cdef char*arr = read_data(buffer, byte_length)
    data_type = numpy.dtype("<i8") if little_endian else numpy.dtype(">i8")
    tag.value_ = numpy.array(numpy.frombuffer(arr[:byte_length], dtype=data_type, count=length), numpy.int64).ravel()
    return tag


cdef class LongArrayTag(AbstractBaseArrayTag):
    """This class behaves like an 1D Numpy signed integer array with each value stored in a long."""
    tag_id = ID_LONG_ARRAY

    def __init__(LongArrayTag self, object value = ()):
        self.value_ = numpy.asarray(value, numpy.int64).ravel()

    cdef str _to_snbt(LongArrayTag self):
        cdef long long elem
        cdef list tags = []
        for elem in self.value_:
            tags.append(f"{elem}L")
        return f"[L;{CommaSpace.join(tags)}]"

    cdef void write_payload(
        LongArrayTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_array(
            numpy.asarray(self.value_, numpy.dtype("<i8") if little_endian else numpy.dtype(">i8")),
            buffer,
            8,
            little_endian
        )

    def __str__(LongArrayTag self):
        return str(self.value_)

    def __reduce__(LongArrayTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(LongArrayTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(LongArrayTag self):
        return self.__class__(self.value_)

    @property
    def np_array(LongArrayTag self):
        """
        A numpy array holding the same internal data.
        Changes to the array will also modify the internal state.
        """
        return self.value_

    def __repr__(LongArrayTag self):
        return f"{self.__class__.__name__}({list(self.value_)})"

    def __eq__(LongArrayTag self, other):
        cdef LongArrayTag other_
        if isinstance(other, LongArrayTag):
            other_ = other
            return numpy.array_equal(self.value_, other_.value_)
        elif __major__ <= 2 and isinstance(other, AbstractBaseArrayTag):
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return numpy.array_equal(self.value_, primitive_conversion(other))
        return NotImplemented

    def __getitem__(LongArrayTag self, item):
        return self.value_.__getitem__(item)

    def __setitem__(LongArrayTag self, key, value):
        self.value_.__setitem__(key, value)

    def __array__(LongArrayTag self, dtype=None):
        return numpy.asarray(self.value_, dtype=dtype)

    def __len__(LongArrayTag self):
        return len(self.value_)

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) + primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) - primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) * primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __matmul__(self, other):
            warnings.warn(f"__matmul__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) @ primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) / primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) // primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) % primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo).astype(numpy.dtype(">i8"))

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) << primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) >> primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) & primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) ^ primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(self) | primitive_conversion(other)).astype(numpy.dtype(">i8"))

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) + primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) - primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) * primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) / primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) // primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) % primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo).astype(numpy.dtype(">i8"))

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) << primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) >> primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) & primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) ^ primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (primitive_conversion(other) | primitive_conversion(self)).astype(numpy.dtype(">i8"))

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (-self.value_).astype(numpy.dtype(">i8"))

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return (+self.value_).astype(numpy.dtype(">i8"))

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on LongArrayTag and will be removed in the future. Please use .np_array to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_).astype(numpy.dtype(">i8"))


from copy import copy
from io import BytesIO
import warnings
import gzip
from typing import Any

from mutf8 import encode_modified_utf8
from ._util cimport write_byte, write_string
from ._dtype import EncoderType


cdef class AbstractBase:
    """Abstract Base class for all Tags and the NamedTag"""

    cpdef str to_snbt(self, object indent=None, object indent_chr=None):
        """
        Return the NBT data in Stringified NBT format.
        
        :param indent: int, str or None. If int will indent with this many spaces. If string will indent with this string. If None will return on one line.
        :param indent_chr: Depreciated. Use indent instead.
        :return: Stringified NBT representation of the data.
        """
        raise NotImplementedError

    def to_nbt(
        self,
        *,
        bint compressed=True,
        bint little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
    ):
        """
        Get the data in binary NBT format.

        :param compressed: Should the bytes be compressed with gzip.
        :param little_endian: Should the bytes be saved in little endian format.
        :param string_encoder: A function to encode strings to bytes.
        :return: The binary NBT representation of the class.
        """
        raise NotImplementedError

    def save_to(
        self,
        object filepath_or_buffer=None,
        *,
        bint compressed=True,
        bint little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
    ):
        """
        Convert the data to the binary NBT format. Optionally write to a file.

        If filepath_or_buffer is a valid file path in string form the data will be written to that file.

        If filepath_or_buffer is a file like object the bytes will be written to it using .write method.

        :param filepath_or_buffer: A path or writeable object to write the data to.
        :param compressed: Should the bytes be compressed with gzip.
        :param little_endian: Should the bytes be saved in little endian format.
        :param string_encoder: A function to encode strings to bytes.
        :return: The binary NBT representation of the class.
        """
        raise NotImplementedError


cdef class AbstractBaseTag(AbstractBase):
    """Abstract Base Class for all Tag classes"""
    tag_id: int = None

    @property
    def py_data(self) -> Any:
        """
        A python representation of the class. Note that the return type is undefined and may change in the future.
        You would be better off using the py_{type} or np_array properties if you require a fixed type.
        This is here for convenience to get a python representation under the same property name.
        """
        raise NotImplementedError

    @property
    def value(self):
        """Legacy property to access the python data. Depreciated. Use :attr:`py_data` or typed py_{type} instead."""
        warnings.warn("value property is depreciated. Use py_{type} instead", DeprecationWarning)
        return self.py_data

    cpdef str to_snbt(self, object indent=None, object indent_chr=None):
        if indent_chr is not None:
            warnings.warn("indent_chr is deprecated. Use indent instead.", DeprecationWarning)
            indent = indent_chr
        if isinstance(indent, int):
            return self._pretty_to_snbt(" " * indent)
        elif isinstance(indent, str):
            if indent.strip():
                raise ValueError("The indent argument must only contain whitespace.")
            return self._pretty_to_snbt(indent)
        return self._to_snbt()

    cdef str _to_snbt(self):
        """Internal method to format the class data as SNBT."""
        raise NotImplementedError

    cdef str _pretty_to_snbt(self, str indent_chr, int indent_count=0, bint leading_indent=True):
        return f"{indent_chr * indent_count * leading_indent}{self._to_snbt()}"

    def to_nbt(
        self,
        *,
        bint compressed=True,
        bint little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
        str name="",
    ):
        """
        Get the data in binary NBT format.

        :param compressed: Should the bytes be compressed with gzip.
        :param little_endian: Should the bytes be saved in little endian format.
        :param string_encoder: A function to encode strings to bytes.
        :param name: The root tag name.
        :return: The binary NBT representation of the class.
        """
        cdef object buffer = BytesIO()
        cdef object gzip_buffer
        self.write_tag(buffer, name, little_endian, string_encoder)
        cdef bytes data = buffer.getvalue()
        if compressed:
            gzip_buffer = BytesIO()
            with gzip.GzipFile(fileobj=gzip_buffer, mode='wb') as gz:
                gz.write(data)
            data = gzip_buffer.getvalue()
        return data

    def save_to(
        self,
        object filepath_or_buffer=None,
        *,
        bint compressed=True,
        bint little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
        str name="",
    ) -> bytes:
        """
        Convert the data to the binary NBT format. Optionally write to a file.

        If filepath_or_buffer is a valid file path in string form the data will be written to that file.

        If filepath_or_buffer is a file like object the bytes will be written to it using .write method.

        :param filepath_or_buffer: A path or writeable object to write the data to.
        :param compressed: Should the bytes be compressed with gzip.
        :param little_endian: Should the bytes be saved in little endian format.
        :param string_encoder: A function to encode strings to bytes.
        :param name: The root tag name.
        :return: The binary NBT representation of the class.
        """
        cdef bytes data = self.to_nbt(
            compressed=compressed,
            little_endian=little_endian,
            string_encoder=string_encoder,
            name=name,
        )

        if filepath_or_buffer is None:
            pass
        elif isinstance(filepath_or_buffer, str):
            with open(filepath_or_buffer, 'wb') as fp:
                fp.write(data)
        else:
            filepath_or_buffer.write(data)
        return data

    cdef void write_tag(
        self,
        object buffer: BytesIO,
        str name,
        bint little_endian,
        string_encoder: EncoderType
    ) except *:
        """
        Write the header and payload to the buffer.
        type_byte, name_len, name, payload
        
        :param buffer: The buffer to write to.
        :param name: The name of the tag in the header.
        :param little_endian: Should the data be written in little endian format.
        :param string_encoder: A function to encode strings to bytes.
        """
        write_byte(self.tag_id, buffer)
        write_string(name, buffer, little_endian, string_encoder)
        self.write_payload(buffer, little_endian, string_encoder)

    cdef void write_payload(
        self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        """
        Write the payload to the buffer.
        This is only the data contained in the class and does not include the header.
        
        :param buffer: The buffer to write to.
        :param little_endian: Should the data be written in little endian format.
        :param string_encoder: A function to encode strings to bytes.
        """
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    def __reduce__(self):
        raise NotImplementedError

    def copy(self):
        """Return a shallow copy of the class"""
        return copy(self)

    def __deepcopy__(self, memo=None):
        raise NotImplementedError

    def __copy__(self):
        raise NotImplementedError


cdef class AbstractBaseImmutableTag(AbstractBaseTag):
    """Abstract Base Class for all immutable Tag classes"""
    def __hash__(self):
        raise NotImplementedError


cdef class AbstractBaseMutableTag(AbstractBaseTag):
    """Abstract Base Class for all mutable Tag classes"""
    __hash__ = None



from setuptools import setup
from Cython.Build import cythonize
import versioneer
import numpy

# Note this will error if no pyx files are present
ext = cythonize(
    f"amulet_nbt/**/*.pyx",
    language_level=3,
)

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    include_dirs=[numpy.get_include()],
    ext_modules=ext,
)



from ._value cimport AbstractBaseImmutableTag


cdef class AbstractBaseNumericTag(AbstractBaseImmutableTag):
    """Abstract Base Class for all numeric Tag classes"""
    def __int__(self):
        raise NotImplementedError

    def __float__(self):
        raise NotImplementedError

    def __bool__(self):
        raise NotImplementedError



from ._value import AbstractBaseImmutableTag

class AbstractBaseNumericTag(AbstractBaseImmutableTag):
    def __int__(self) -> int:
        raise NotImplementedError
    def __float__(self) -> float:
        raise NotImplementedError
    def __bool__(self) -> bool:
        raise NotImplementedError



import gzip
import warnings
import zlib
from io import BytesIO
from typing import Union, BinaryIO, List, Tuple
import os

from mutf8 import decode_modified_utf8

from ._errors import NBTLoadError
from ._util cimport BufferContext, read_string, read_byte
from ._value cimport AbstractBaseTag
from ._int cimport (
    read_byte_tag,
    read_short_tag,
    read_int_tag,
    read_long_tag,
)
from ._float cimport (
    read_float_tag,
    read_double_tag,
)
from ._array cimport (
    read_byte_array_tag,
    read_int_array_tag,
    read_long_array_tag,
)
from ._string cimport read_string_tag
from ._list cimport read_list_tag
from ._compound cimport read_compound_tag
from ._named_tag cimport NamedTag
from ._dtype import DecoderType


cpdef inline bytes safe_gunzip(bytes data):
    if data[:2] == b'\x1f\x8b':  # if the first two bytes are this it should be gzipped
        try:
            data = gzip.GzipFile(fileobj=BytesIO(data)).read()
        except (IOError, zlib.error) as e:
            pass
    return data


cdef BufferContext get_buffer(
    object filepath_or_buffer: Union[str, bytes, BinaryIO, memoryview, None],
    bint compressed=True,
):
    cdef bytes data
    if isinstance(filepath_or_buffer, str):
        # if a string load from the file path
        if not os.path.isfile(filepath_or_buffer):
            raise NBTLoadError(f"There is no file at {filepath_or_buffer}")
        with open(filepath_or_buffer, "rb") as f:
            data = f.read()
    elif isinstance(filepath_or_buffer, bytes):
        data = filepath_or_buffer
    elif isinstance(filepath_or_buffer, memoryview):
        data = filepath_or_buffer.tobytes()
    elif hasattr(filepath_or_buffer, "read"):
        buffer_data = filepath_or_buffer.read()
        if not isinstance(buffer_data, bytes):
            raise NBTLoadError(f"buffer.read() must return a bytes object. Got {type(buffer_data)} instead.")
        data = buffer_data
    else:
        raise NBTLoadError("buffer did not have a read method.")

    if compressed:
        data = safe_gunzip(data)
    return BufferContext(data)


cdef AbstractBaseTag load_payload(BufferContext buffer, char tag_type, bint little_endian, string_decoder: DecoderType):
    if tag_type == 1:
        return read_byte_tag(buffer, little_endian)
    elif tag_type == 2:
        return read_short_tag(buffer, little_endian)
    elif tag_type == 3:
        return read_int_tag(buffer, little_endian)
    elif tag_type == 4:
        return read_long_tag(buffer, little_endian)
    elif tag_type == 5:
        return read_float_tag(buffer, little_endian)
    elif tag_type == 6:
        return read_double_tag(buffer, little_endian)
    elif tag_type == 7:
        return read_byte_array_tag(buffer, little_endian)
    elif tag_type == 8:
        return read_string_tag(buffer, little_endian, string_decoder)
    elif tag_type == 9:
        return read_list_tag(buffer, little_endian, string_decoder)
    elif tag_type == 10:
        return read_compound_tag(buffer, little_endian, string_decoder)
    elif tag_type == 11:
        return read_int_array_tag(buffer, little_endian)
    elif tag_type == 12:
        return read_long_array_tag(buffer, little_endian)
    else:
        raise NBTLoadError(f"Tag {tag_type} does not exist.")


cdef inline tuple load_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType):
    cdef char tag_type = read_byte(buffer)
    cdef str name = read_string(buffer, little_endian, string_decoder)
    return name, load_payload(buffer, tag_type, little_endian, string_decoder)


cdef NamedTag load_named_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType):
    cdef str name
    cdef AbstractBaseTag tag
    name, tag = load_tag(buffer, little_endian, string_decoder)
    return NamedTag(tag, name)


cdef class ReadContext:
    def __cinit__(self):
        self.offset = 0


def _load_one(
    object filepath_or_buffer: Union[str, bytes, BinaryIO, memoryview, None],
    *,
    bint compressed: bool=True,
    bint little_endian: bool = False,
    ReadContext read_context = None,
    string_decoder: DecoderType = decode_modified_utf8
) -> NamedTag:
    cdef BufferContext buffer = get_buffer(filepath_or_buffer, compressed)
    if buffer.size < 1:
        raise EOFError("buffer is empty")
    cdef NamedTag tag = load_named_tag(buffer, little_endian, string_decoder)
    if read_context is not None:
        read_context.offset = buffer.offset
    return tag


def load_many(
    object filepath_or_buffer: Union[str, bytes, BinaryIO, memoryview, None],
    *,
    int count = 1,
    bint compressed: bool=True,
    bint little_endian: bool = False,
    ReadContext read_context = None,
    string_decoder: DecoderType = decode_modified_utf8
) -> List[NamedTag]:
    if count < 1:
        raise ValueError("Count must be 1 or more.")
    cdef BufferContext buffer = get_buffer(filepath_or_buffer, compressed)
    if buffer.size < 1:
        raise EOFError("buffer is empty")
    cdef list results = []
    cdef size_t i
    for i in range(count):
        results.append(
            load_named_tag(buffer, little_endian, string_decoder)
        )
    if read_context is not None:
        read_context.offset = buffer.offset
    return results


def load(
    object filepath_or_buffer: Union[str, bytes, BinaryIO, memoryview, None],
    *args,
    bint compressed: bool=True,
    object count: int = None,
    bint offset: bool = False,
    bint little_endian: bool = False,
    string_decoder: DecoderType = decode_modified_utf8,
    ReadContext read_context = None,
) -> Union[NamedTag, Tuple[Union[NamedTag, List[NamedTag]], int]]:
    if args:
        warnings.warn("load arguments are going to be keyword only in the future. This function passes the inputs positionally.", FutureWarning)
        compressed, *args = args
        if args:
            count, *args = args
        if args:
            offset, *args = args
        if args:
            little_endian, *args = args

    if offset and read_context is None:
        warnings.warn("load with offset is depreciated. In future versions this function will only return the result. To get the offset pass a ReadContext instance to read_context input.", FutureWarning)
        read_context = ReadContext()

    cdef object result

    if count is None:
        result = _load_one(
            filepath_or_buffer,
            compressed=compressed,
            little_endian=little_endian,
            read_context=read_context,
            string_decoder=string_decoder
        )
    else:
        warnings.warn("load with count is depreciated. Use load_many", DeprecationWarning)
        result = load_many(
            filepath_or_buffer,
            count=count,
            compressed=compressed,
            little_endian=little_endian,
            read_context=read_context,
            string_decoder=string_decoder
        )

    if offset:
        # depreciated
        return result, read_context.offset
    else:
        return result



cdef str CommaNewline = ",\n"
cdef str CommaSpace = ", "

cdef char ID_END = 0
cdef char ID_BYTE = 1
cdef char ID_SHORT = 2
cdef char ID_INT = 3
cdef char ID_LONG = 4
cdef char ID_FLOAT = 5
cdef char ID_DOUBLE = 6
cdef char ID_BYTE_ARRAY = 7
cdef char ID_STRING = 8
cdef char ID_LIST = 9
cdef char ID_COMPOUND = 10
cdef char ID_INT_ARRAY = 11
cdef char ID_LONG_ARRAY = 12
cdef char ID_MAX = 12
